var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import prompts from 'prompts';
import { Command } from 'commander';
import { createRecordsClient, } from '@casual-simulation/aux-records/RecordsClient';
import { askForInputs } from './schema';
import repl from 'node:repl';
// @ts-ignore
import Conf from 'conf';
import { parseSessionKey } from '@casual-simulation/aux-records/AuthUtils';
import { PassThrough } from 'node:stream';
const REFRESH_LIFETIME_MS = 1000 * 60 * 60 * 24 * 7; // 1 week
const config = new Conf({
    projectName: 'casualos-cli',
});
const program = new Command();
program
    .name('casualos')
    .description('A CLI for CasualOS')
    .version(GIT_TAG)
    .option('-e, --endpoint <url>', 'The endpoint to use for queries. Can be used to override the current endpoint.');
program
    .command('login')
    .description('Login to the CasualOS API')
    .action(() => __awaiter(void 0, void 0, void 0, function* () {
    const opts = program.optsWithGlobals();
    const endpoint = yield getEndpoint(opts.endpoint);
    const client = yield getClient(endpoint, getSessionKey(endpoint));
    yield login(client);
}));
program
    .command('logout')
    .description('Logout of the CasualOS API')
    .action(() => __awaiter(void 0, void 0, void 0, function* () {
    const opts = program.optsWithGlobals();
    const endpoint = yield getEndpoint(opts.endpoint);
    saveSessionKey(endpoint, null);
    console.log('Logged out!');
}));
program
    .command('set-endpoint')
    .argument('[endpoint]', 'The endpoint to use for queries. If omitted, then you will be prompted to enter an endpoint.')
    .description('Set the endpoint that is currently in use.')
    .action((endpoint) => __awaiter(void 0, void 0, void 0, function* () {
    if (endpoint) {
        saveCurrentEndpoint(endpoint);
    }
    else {
        endpoint = yield updateEndpoint();
    }
    const origin = getOrigin(endpoint);
    if (origin !== endpoint) {
        const response = yield prompts({
            type: 'confirm',
            name: 'confirm',
            message: `The origin for the endpoint (${origin}) is different from the endpoint (${endpoint}) itself. Do you want to keep it?`,
            initial: true,
        });
        if (response.confirm) {
            return;
        }
    }
    const originResponse = yield prompts({
        type: 'text',
        name: 'origin',
        message: 'Enter the origin to use for requests to this endpoint.',
        initial: endpoint,
    });
    saveOrigin(endpoint, originResponse.origin);
}));
program
    .command('status')
    .description('Get the status of the current session.')
    .action(() => __awaiter(void 0, void 0, void 0, function* () {
    const endpoint = getCurrentEndpoint();
    printStatus(endpoint);
}));
program
    .command('query')
    .description('Query the CasualOS API')
    .argument('[procedure]', 'The procedure to execute. If omitted, then you will be prompted to select a procedure.')
    .argument('[input]', 'The input to the procedure. If specified, then it will be parsed as JSON. If omitted, then you will be prompted to enter the input.')
    .option('-k, --key <key>', 'The session key to use for the query. If not specified, then the current session key will be used.')
    .action((procedure, input, options) => __awaiter(void 0, void 0, void 0, function* () {
    var _a;
    const opts = program.optsWithGlobals();
    const endpoint = yield getEndpoint(opts.endpoint);
    const client = yield getClient(endpoint, (_a = opts.key) !== null && _a !== void 0 ? _a : (yield getOrRefreshSessionKey(endpoint)));
    yield query(client, procedure, input);
}));
program
    .command('repl')
    .description('Start a REPL for the CasualOS API')
    .option('-k, --key <key>', 'The session key to use for the session. If omitted, then the current session key will be used.')
    .addHelpText('after', `\nThe CasualOS REPL allows you to interact with the CasualOS API using a Read-Eval-Print Loop (REPL).\nIt supports JavaScript and has a special function, query([procedure], [input]), that can be used to query the API.`)
    .action((options) => __awaiter(void 0, void 0, void 0, function* () {
    var _b;
    const opts = program.optsWithGlobals();
    const endpoint = yield getEndpoint(opts.endpoint);
    const client = yield getClient(endpoint, (_b = opts.key) !== null && _b !== void 0 ? _b : (yield getOrRefreshSessionKey(endpoint)));
    const replIn = new PassThrough();
    process.stdin.pipe(replIn);
    function pauseRepl(func) {
        return (...args) => __awaiter(this, void 0, void 0, function* () {
            process.stdin.unpipe(replIn);
            replIn.pause();
            try {
                return yield func(...args);
            }
            finally {
                replIn.resume();
                process.stdin.pipe(replIn);
            }
        });
    }
    const replServer = repl.start({
        prompt: 'casualos > ',
        input: replIn,
        output: process.stdout,
    });
    replServer.on('exit', () => {
        process.stdin.unpipe(replIn);
    });
    Object.defineProperties(replServer.context, {
        query: {
            configurable: false,
            writable: false,
            enumerable: true,
            value: pauseRepl((procedure, input) => __awaiter(void 0, void 0, void 0, function* () {
                return yield query(client, procedure, input, false, true, replServer);
            })),
        },
    });
}));
function query(client, procedure, input, shouldConfirm = true, isJavaScriptInput = false, repl = null) {
    return __awaiter(this, void 0, void 0, function* () {
        const availableOperations = yield client.listProcedures({});
        while (!procedure ||
            !availableOperations.procedures.find((p) => p.name === procedure)) {
            const response = yield prompts({
                type: 'autocomplete',
                name: 'procedure',
                message: 'Select the procedure to execute',
                choices: availableOperations.procedures.map((op) => ({
                    title: op.name,
                    value: op.name,
                })),
            });
            procedure = response.procedure;
        }
        const operation = availableOperations.procedures.find((p) => p.name === procedure);
        if (!operation) {
            console.error(`Could not find operation ${procedure}!`);
            return;
        }
        console.log('Your selected operation:', operation);
        if (!input) {
            input = yield askForInputs(operation.inputs, operation.name, repl);
        }
        else if (!isJavaScriptInput) {
            input = JSON.parse(input);
        }
        console.log('Your input:', input);
        let continueRequest = true;
        if (shouldConfirm) {
            const confirm = yield prompts({
                type: 'confirm',
                name: 'continue',
                message: 'Do you want to continue?',
                initial: true,
            });
            continueRequest = confirm.continue;
        }
        if (continueRequest) {
            const result = yield callProcedure(client, operation.name, input);
            if (shouldConfirm) {
                console.log('Result:', result);
            }
            return result;
        }
        else {
            console.log('Cancelled.');
        }
    });
}
function callProcedure(client, operation, input) {
    return __awaiter(this, void 0, void 0, function* () {
        while (true) {
            const result = yield client.callProcedure(operation, input, {
                headers: getHeaders(client),
            });
            if (result.success === false && result.errorCode === 'not_logged_in') {
                const loginResponse = yield prompts({
                    type: 'confirm',
                    name: 'login',
                    message: 'You are not logged in. Do you want to log in and try again?',
                    initial: true,
                });
                if (loginResponse.login) {
                    const key = yield login(client);
                    if (!key) {
                        return result;
                    }
                }
                else {
                    return result;
                }
            }
            else if (result.success === false &&
                result.errorCode === 'invalid_origin') {
                const originResponse = yield prompts({
                    type: 'text',
                    name: 'origin',
                    message: 'The endpoint does not allow itself as an origin. Enter the origin to use for the request.',
                });
                saveOrigin(client.endpoint, originResponse.origin);
            }
            else {
                return result;
            }
        }
    });
}
function getClient(endpoint, key) {
    return __awaiter(this, void 0, void 0, function* () {
        printStatus(endpoint);
        const client = createRecordsClient(endpoint);
        if (key) {
            client.sessionKey = key;
        }
        return client;
    });
}
function getOrRefreshSessionKey(endpoint) {
    return __awaiter(this, void 0, void 0, function* () {
        const key = getSessionKey(endpoint);
        if (isExpired(key)) {
            return null;
        }
        else if (willExpire(key)) {
            return yield replaceSessionKey(endpoint, key);
        }
        return key;
    });
}
function getSessionKey(endpoint) {
    return convertToString(config.get(`${endpoint}:sessionKey`));
}
function saveSessionKey(endpoint, key) {
    config.set(`${endpoint}:sessionKey`, key);
}
function saveLoginResult(client, result) {
    if (!result) {
        saveSessionKey(client.endpoint, null);
        client.sessionKey = null;
    }
    else {
        saveSessionKey(client.endpoint, result.sessionKey);
        client.sessionKey = result.sessionKey;
    }
}
function getExpiration(key) {
    const parsed = parseSessionKey(key);
    if (!parsed) {
        return 0;
    }
    const [userId, sessionId, secret, expireTimeMs] = parsed;
    return expireTimeMs;
}
function isExpired(key) {
    const expireTimeMs = getExpiration(key);
    return expireTimeMs < Date.now();
}
function willExpire(key) {
    const expireTimeMs = getExpiration(key);
    const lifetimeMs = expireTimeMs - Date.now();
    const refreshTimeMs = Math.max(lifetimeMs - REFRESH_LIFETIME_MS, 0);
    return refreshTimeMs < 0;
}
function login(client) {
    return __awaiter(this, void 0, void 0, function* () {
        const response = yield prompts([
            {
                type: 'select',
                name: 'type',
                message: 'Select the type of address to use for login.',
                choices: [
                    {
                        title: 'Email',
                        value: 'email',
                    },
                    {
                        title: 'Phone',
                        value: 'phone',
                    },
                    {
                        title: 'Privo',
                        value: 'privo',
                    },
                ],
            },
        ]);
        if (response.type === 'privo') {
            return yield loginWithPrivo(client);
        }
        else if (response.type === 'email' || response.type === 'phone') {
            const addressResponse = yield prompts({
                type: 'text',
                name: 'address',
                message: 'Enter your address.',
            });
            const addressType = response.type;
            const address = addressResponse.address;
            return yield loginWithCode(client, address, addressType);
        }
    });
}
function loginWithCode(client, address, addressType) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = yield client.requestLogin({
            address: address,
            addressType: addressType,
        }, {
            headers: getHeaders(client),
        });
        if (result.success) {
            const response = yield prompts({
                type: 'text',
                name: 'code',
                message: 'Enter the code that was sent to your address.',
            });
            const code = response.code;
            const loginResult = yield client.completeLogin({
                code: code,
                requestId: result.requestId,
                userId: result.userId,
            }, {
                headers: getHeaders(client),
            });
            if (loginResult.success === true) {
                saveLoginResult(client, loginResult);
                console.log('Login successful!');
                return loginResult.sessionKey;
            }
            else {
                saveLoginResult(client, null);
                console.log('Failed to complete login:');
                console.log(loginResult);
                return null;
            }
        }
        else {
            saveLoginResult(client, null);
            console.log('Failed to create login request:');
            console.log(result);
            return null;
        }
    });
}
function loginWithPrivo(client) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = yield client.requestPrivoLogin({}, {
            headers: getHeaders(client),
        });
        if (result.success === false) {
            saveLoginResult(client, null);
            console.log('Failed to request Privo login:');
            console.log(result);
            return null;
        }
        const open = (yield import('open')).default;
        yield open(result.authorizationUrl);
        const startTime = Date.now();
        const timeout = 1000 * 60 * 5; // 5 minutes
        while (Date.now() - startTime < timeout) {
            const loginResult = yield client.completeOAuthLogin({
                requestId: result.requestId,
            }, {
                headers: getHeaders(client),
            });
            if (loginResult.success === true) {
                saveLoginResult(client, loginResult);
                console.log('Login successful!');
                return loginResult.sessionKey;
            }
            else {
                if (loginResult.errorCode === 'not_completed') {
                    // Wait for a second before trying again.
                    yield new Promise((resolve) => setTimeout(resolve, 1000));
                }
                else {
                    saveLoginResult(client, null);
                    console.log('Failed to complete login:');
                    console.log(loginResult);
                    return null;
                }
            }
        }
        saveLoginResult(client, null);
        console.log('Failed to complete login:');
        console.log('Timed out');
        return null;
    });
}
function replaceSessionKey(endpoint, key) {
    return __awaiter(this, void 0, void 0, function* () {
        const client = yield getClient(endpoint, key);
        const result = yield client.replaceSession(undefined, {
            sessionKey: key,
        });
        if (result.success === true) {
            saveSessionKey(endpoint, result.sessionKey);
            console.log('Session key replaced!');
            return result.sessionKey;
        }
        saveSessionKey(endpoint, null);
        console.log('Failed to replace session key:');
        console.log(result);
        return null;
    });
}
function getEndpoint(endpoint) {
    return __awaiter(this, void 0, void 0, function* () {
        if (endpoint) {
            return endpoint;
        }
        let savedEndpoint = getCurrentEndpoint();
        console.log('saved endpoint', savedEndpoint);
        if (!savedEndpoint) {
            savedEndpoint = yield updateEndpoint();
        }
        return savedEndpoint;
    });
}
function getCurrentEndpoint() {
    return convertToString(config.get('currentEndpoint'));
}
function updateEndpoint() {
    return __awaiter(this, void 0, void 0, function* () {
        const response = yield prompts({
            type: 'text',
            name: 'endpoint',
            message: 'Enter the endpoint to use for queries.',
        });
        const savedEndpoint = response.endpoint;
        saveCurrentEndpoint(savedEndpoint);
        return savedEndpoint;
    });
}
function saveCurrentEndpoint(endpoint) {
    config.set('currentEndpoint', endpoint);
    console.log('Endpoint updated to:', endpoint);
}
function saveOrigin(endpoint, origin) {
    config.set(`${endpoint}:origin`, origin);
}
function getOrigin(endpoint) {
    let origin = config.get(`${endpoint}:origin`);
    if (typeof origin === 'string' && origin) {
        return origin;
    }
    return endpoint;
}
function getHeaders(client) {
    return {
        origin: getOrigin(client.endpoint),
    };
}
function printStatus(endpoint) {
    if (!endpoint) {
        console.log('No endpoint is currently set.');
        console.log('Set an endpoint using the set-endpoint command.');
        return;
    }
    console.log('Current endpoint:', endpoint);
    const key = getSessionKey(endpoint);
    if (key) {
        const expire = getExpiration(key);
        if (expire < Date.now()) {
            console.log('The current session has expired.');
        }
        else {
            console.log('You are logged in.');
            console.log('Session expires at:', new Date(expire).toString());
        }
    }
    else {
        console.log('You are not logged in.');
    }
}
function main() {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield program.parseAsync(process.argv);
        }
        catch (err) {
            console.error(err);
        }
    });
}
function convertToString(str) {
    if (typeof str === 'undefined' || str === null) {
        return str;
    }
    return String(str);
}
main();
//# sourceMappingURL=cli.js.map